<?php
if (file_exists( 'config/config.php')) {
    require_once 'config/config.php';
}

define( 'CLR_BLACK',       "\033[0;30m");
define( 'CLR_DARKGRAY',    "\033[1;30m");
define( 'CLR_BLUE',        "\033[0;34m");
define( 'CLR_LIGHTBLUE',   "\033[1;34m");
define( 'CLR_GREEN',       "\033[0;32m");
define( 'CLR_LIGHTGREEN',  "\033[1;32m");
define( 'CLR_CYAN',        "\033[0;36m");
define( 'CLR_LIGHTCYAN',   "\033[1;36m");
define( 'CLR_RED',         "\033[0;31m");
define( 'CLR_LIGHTRED',    "\033[1;31m");
define( 'CLR_PURPLE',      "\033[0;35m");
define( 'CLR_LIGHTPURPLE', "\033[1;35m");
define( 'CLR_BROWN',       "\033[0;33m");
define( 'CLR_YELLOW',      "\033[1;33m");
define( 'CLR_LIGHTGRAY',   "\033[0;37m");
define( 'CLR_WHITE',       "\033[1;37m");
define( 'CLR_NONE',        "\033[0m");

class TaskManager {

    private static $_instance = null;
    private $_task = 'default';
    private $_args = null;

    private function __construct() {
        $this->arguments();
    }

    public static function run() {
         self::getInstance()->getTask()->run();
    }

    public function getTask() {
        if (!file_exists( 'config/config.php')) {
            echo CLR_YELLOW . "Looks like testing is not confugured yet.\nWhould you like to configure it now? [Y/n]: " . CLR_NONE;
            $answer = strtolower( trim( fgets( STDIN)));
            if (!$answer || $answer == 'y' || $answer == 'yes') {
                require_once 'lib/task/Configure.php';
                return new ConfigureTask();
            }
        }
        if (($task = self::getArg( 0))) {
            $this->_task = $task;
        }
        $task = ucfirst( $this->_task);
        $file = "lib/task/$task.php";
        if (file_exists( $file)) {
            require_once $file;
            $class = $task . 'Task';
            return new $class();
        } else {
            echo CLR_LIGHTRED . 'Cannot find specified task: "' . $this->_task . '". Aborted!' . "\n" . CLR_NONE;
            require_once 'lib/task/Default.php';
            return new DefaultTask();
        }
    }

    public static function getInstance() {
        if (is_null( self::$_instance)) {
            self::$_instance = new self();
        }
        return self::$_instance;
    }

    public static function getScript() {
        $args = self::getInstance()->arguments();
        return $args['exec'];
    }

    public static function flagIsset( $flag) {
        $args = self::getInstance()->arguments();
        for ($i = 0, $s = sizeof( $args['flags']); $i < $s; $i++) {
            if ($args['flags'][$i] == $flag) {
                return true;
            }
        }
        return false;
    }

    public static function getArg( $indx) {
        $args = self::getInstance()->arguments();
        return @$args['arguments'][$indx];
    }

    public function arguments() {
        if (!is_null( $this->_args)) {
            return $this->_args;
        }

        $args = $_SERVER['argv'];

        $ret = array(
            'exec'      => '',
            'options'   => array(),
            'flags'     => array(),
            'arguments' => array(),
        );

        $ret['exec'] = array_shift( $args );

        while (($arg = array_shift($args)) != NULL) {
            // Is it a option? (prefixed with --)
            if ( substr($arg, 0, 2) === '--' ) {
                $option = substr($arg, 2);

                // is it the syntax '--option=argument'?
                if (strpos($option,'=') !== FALSE)
                    array_push( $ret['options'], explode('=', $option, 2) );
                else
                    array_push( $ret['options'], $option );

                continue;
            }

            // Is it a flag or a serial of flags? (prefixed with -)
            if ( substr( $arg, 0, 1 ) === '-' ) {
                for ($i = 1; isset($arg[$i]) ; $i++)
                    $ret['flags'][] = $arg[$i];

                continue;
            }

            // finally, it is not option, nor flag
            $ret['arguments'][] = $arg;
            continue;
        }
        $this->_args = $ret;
        return $this->_args;
    }

}
