<?php
require_once 'Interface.php';
require_once 'db/Abstract.php';

class BuildTask extends Db implements TaskInterface {

    private $_tables = array(
          'al_tree' => array( false, 'CREATE TABLE `al_tree` (`id` bigint(20) NOT NULL auto_increment, `parent_id` bigint(20) default NULL, `name` varchar(50) NOT NULL, PRIMARY KEY  (`id`), KEY `fk_tree_tree` (`parent_id`), CONSTRAINT `fk_tree_tree` FOREIGN KEY (`parent_id`) REFERENCES `al_tree` (`id`) ON DELETE CASCADE ON UPDATE CASCADE) ENGINE=InnoDB AUTO_INCREMENT=100009 DEFAULT CHARSET=utf8')
        , 'ns_tree' => array( false, 'CREATE TABLE `ns_tree` (`id` bigint(20) NOT NULL auto_increment, `name` varchar(50) NOT NULL, `lft` bigint(20) NOT NULL, `rgt` bigint(20) NOT NULL, `level` bigint(20) NOT NULL, PRIMARY KEY  (`id`), KEY `nslrl_idx` (`lft`,`rgt`,`level`)) ENGINE=InnoDB AUTO_INCREMENT=100009 DEFAULT CHARSET=utf8')
        , 'mp_tree' => array( false, 'CREATE TABLE `mp_tree` (`id` bigint(20) NOT NULL auto_increment, `name` varchar(50) NOT NULL, `path` varchar(100) NOT NULL, `level` INTEGER NOT NULL,  PRIMARY KEY  (`id`), KEY `mpp_idx` (`path`)) ENGINE=InnoDB DEFAULT CHARSET=utf8')
    );

    public function run() {
        echo CLR_YELLOW . "Build task started...\n" . CLR_NONE;
        $start = microtime( true);
        $this->_fixDbTables();
        $this->_fixDbFunctions();
        $this->_buildAlTree();
        $this->_buildNsTree();
        $this->_buildMpTree();
        $time = number_format( microtime( true) - $start, 2, '.', ' ');
        echo CLR_YELLOW . "Build task finished in $time seconds!\n" . CLR_NONE;
    }

    private function _fixDbFunctions() {
        echo CLR_WHITE . "Creating database functions..." . CLR_NONE;
        $this->_query( 'DROP FUNCTION `STRFIND`');
        $this->_query( '--
-- Функция ищет количество вхождений символа delimtr
-- в строку str.
-- Фактически эта функция используется для поиска
-- уровня вложенности узла в дереве типа
-- MATERIALIZED PATH
-- (Кол-во разделителей в пути узла и есть ур-нь.
-- вложенности)
--
-- @param str VARCHAR(255) - исходная строка
-- @param delimtr CHAR(1) - искомый символ
-- @return INT - кол-во вхождений
--
CREATE FUNCTION `STRFIND`(str VARCHAR(255), delimtr CHAR(1)) RETURNS INT
BEGIN
  DECLARE _cnt INT;
  DECLARE _start INT;
  SET _cnt = -1;
  SET _start = 1;
  WHILE _start > 0 DO
    SET _start = LOCATE( delimtr, str);
    SET str    = SUBSTRING( str, _start + 1);
    SET _cnt   = _cnt + 1;
  END WHILE;
  RETURN _cnt;
END');
        if (mysql_errno()) {
            echo "\n" . CLR_LIGHTRED . 'DATABASE ERROR: ' . mysql_error() . CLR_NONE . "\n";
            exit(0);
        }
        $this->_query( 'DROP FUNCTION `REPLACE_PATH`');
        $this->_query( '--
-- Функция замещает в заданной строке _str
-- подстроку _match на строку _replace,.
-- если _match найдена начиная от начала строки _str
-- Удобна для использования в деревьях типа
-- MATERIALIZED PATH для изменения путей
--
-- @param _str VARCHAR(255) - исходная строка
-- @param _match VARCHAR(255) - подстрока поиска
-- @param _replace VARCHAR(255) - подстрока замены
-- @return VARCHAR(255) - новыя строка
--
CREATE FUNCTION `REPLACE_PATH`( _str VARCHAR(255), _match VARCHAR(255), _replace VARCHAR(255)) RETURNS VARCHAR(255)
BEGIN
  IF _str LIKE CONCAT(_match, \'%\') THEN
    RETURN CONCAT( _replace, SUBSTRING( _str, LENGTH(_match)+1, LENGTH(_str)));
  END IF;
  RETURN _str;
END');
        if (mysql_errno()) {
            echo "\n" . CLR_LIGHTRED . 'DATABASE ERROR: ' . mysql_error() . CLR_NONE . "\n";
            exit(0);
        }

        $this->_query( 'DROP PROCEDURE `MOVE_NODE_NS`');
        $this->_query( '--
-- ПРОЦЕДУРА ПЕРЕМЕЩЕНИЯ ВЕТКИ В ДЕРЕВЕ ТИПА NESTED SET
--
-- @param node_id - идентификатор узла, который следует переместить
-- @param parent_id - идентификатор узла в который следует переместить
--
CREATE PROCEDURE MOVE_NODE_NS( node_id BIGINT, parent_id BIGINT)
BEGIN
  DECLARE done INT DEFAULT 0;
  DECLARE c_id, c_lft, c_rgt, c_lvl, nWidth, nLeft, nRight, dtKey, nLvl, pRight, addLvl, addKey BIGINT;
  DECLARE c_name VARCHAR(50);

  -- Создаем курсор который будет хранить всю ветку,
  -- которую мы перемещаем
  DECLARE mvBranch CURSOR FOR
    SELECT id, name, lft - dtKey, rgt - dtKey, level - nLvl FROM ns_tree
    WHERE lft >= nLeft AND rgt <= nRight;

  DECLARE CONTINUE HANDLER FOR NOT FOUND SET done = 1;

  -- Собираем необходимые данные о выбранной ветке
  SELECT rgt - lft + 1, lft, rgt, lft - 1, level INTO nWidth, nLeft, nRight, dtKey, nLvl
  FROM ns_tree WHERE id = node_id;

  -- Выбираем ветку в курсор
  OPEN mvBranch;

  -- Удаляем ветку из дерева
  DELETE FROM ns_tree WHERE lft BETWEEN nLeft AND nRight;

  -- Обновляем ключи в дереве
  UPDATE ns_tree SET rgt = rgt - nWidth WHERE rgt > nRight;
  UPDATE ns_tree SET lft = lft - nWidth WHERE lft > nRight;

  -- выбираем данные о родителе в новом дереве
  SELECT rgt, level + 1 INTO pRight, addLvl FROM ns_tree WHERE id = parent_id;

  SELECT MAX(node.rgt) INTO addKey FROM ns_tree node, ns_tree parent
  WHERE node.lft BETWEEN parent.lft AND parent.rgt AND node.level = parent.level + 1 AND parent.id = parent_id;

  -- создаем в дереве пространство, куда будет
  -- помещена выбранная ранее ветка
  UPDATE ns_tree SET rgt = rgt + nWidth WHERE rgt >= pRight;
  UPDATE ns_tree SET lft = lft + nWidth WHERE lft > pRight;

  -- преобразуем ветку должным образом и вставляем.
  -- в новое дерево
  REPEAT
    FETCH mvBranch INTO c_id, c_name, c_lft, c_rgt, c_lvl;
    IF NOT done THEN
      INSERT INTO ns_tree VALUES (c_id, c_name, c_lft + addKey, c_rgt + addKey, c_lvl + addLvl);
    END IF;
  UNTIL done END REPEAT;

  CLOSE mvBranch;
END');
        if (mysql_errno()) {
            echo "\n" . CLR_LIGHTRED . 'DATABASE ERROR: ' . mysql_error() . CLR_NONE . "\n";
            exit(0);
        }

        $this->_query( 'DROP PROCEDURE `MOVE_NODE_MP`');
        $this->_query( '--
-- ПРОЦЕДУРА ПЕРЕМЕЩЕНИЯ ВЕТКИ В ДЕРЕВЕ ТИПА MATERIALIZED PATH
--
-- @param node_id - идентификатор узла, который следует переместить
-- @param parent_id - идентификатор узла в который следует переместить
--
CREATE PROCEDURE MOVE_NODE_MP( node_id BIGINT, parent_id BIGINT)
BEGIN
  DECLARE done, m_cnt, m_rows, p_cnt, p_rows INT DEFAULT 0;
  DECLARE c_id, p_id, n_pos, n_lvl, np_id, np_lvl, new_pos, dt_lvl, ch_id, ch_pos BIGINT;
  DECLARE c_path, p_path, n_path, np_path, ch_path, new_path VARCHAR(100);

  -- Создаем курсор, в который выбираем ветку,
  -- которую следует переместить
  DECLARE mvBranch CURSOR FOR
    SELECT SQL_CALC_FOUND_ROWS node.id, node.path FROM mp_tree node, mp_tree parent
    WHERE node.path LIKE CONCAT(parent.path, \'%\') AND parent.id = node_id;

  -- Создаем курсор, в который вибираем все ветки справа
  -- от перемещаемой, находящиеся с ней на одном уровне
  DECLARE pChildren CURSOR FOR
    SELECT SQL_CALC_FOUND_ROWS node.id, node.path,
      CAST(SUBSTRING(REVERSE(node.path), 1, LOCATE(\'.\', node.path)-1) AS UNSIGNED) as pos
      FROM mp_tree node, mp_tree parent
      WHERE node.path LIKE CONCAT(parent.path, \'%\') AND node.level = parent.level + 1
      AND CAST(SUBSTRING(REVERSE(node.path), 1, LOCATE(\'.\', node.path)-1) AS UNSIGNED) > n_pos
      AND parent.id = p_id
    ORDER BY pos;

  DECLARE CONTINUE HANDLER FOR NOT FOUND SET done = 1;

  -- Собираем необходимые данные
  SELECT path, CAST(SUBSTRING(REVERSE(path), 1, LOCATE(\'.\', path)-1) AS UNSIGNED), level INTO n_path, n_pos, n_lvl FROM mp_tree
  WHERE id = node_id;

  SELECT parent.id, parent.path INTO p_id, p_path FROM mp_tree node, mp_tree parent
  WHERE parent.path = SUBSTRING( node.path, 1, (LENGTH(node.path) - LOCATE(\'.\', REVERSE(node.path))))
  AND node.id = node_id;

  SELECT id, path, level INTO np_id, np_path, np_lvl FROM mp_tree WHERE id = parent_id;

  -- Находим разницу в уровнях между уровнями перемещаемых узлов
  -- и новым родителем:
  SET dt_lvl = np_lvl - n_lvl + 1;

  SELECT MAX(CAST(SUBSTRING(REVERSE(node.path), 1, LOCATE(\'.\', node.path)-1) AS UNSIGNED)) + 1
  INTO new_pos FROM mp_tree node, mp_tree parent WHERE node.path LIKE CONCAT(parent.path, \'%\')
  AND node.level = parent.level + 1 AND parent.id = parent_id;

  -- Перемещаем ветку
  OPEN mvBranch;
  SELECT FOUND_ROWS() INTO m_rows;

  WHILE m_cnt < m_rows DO
    FETCH mvBranch INTO c_id, c_path;
    UPDATE mp_tree
    SET path  = REPLACE_PATH(path, n_path, CONCAT(np_path, \'.\', new_pos)), level = level + dt_lvl WHERE id = c_id;
    SET m_cnt = m_cnt + 1;
  END WHILE;
  CLOSE mvBranch;

  -- Исправляем данные о путях в ветке из которой.
  -- переместили заданную ветку
  OPEN pChildren;
  SELECT FOUND_ROWS() INTO p_rows;

  WHILE p_cnt < p_rows DO
    FETCH pChildren INTO ch_id, ch_path, ch_pos;
    UPDATE mp_tree SET path = REPLACE_PATH(path, ch_path, CONCAT(p_path, \'.\', ch_pos - 1)) WHERE path LIKE CONCAT( ch_path, \'%\');
    SET p_cnt = p_cnt + 1;
  END WHILE;
  CLOSE pChildren;
END');
        if (mysql_errno()) {
            echo "\n" . CLR_LIGHTRED . 'DATABASE ERROR: ' . mysql_error() . CLR_NONE . "\n";
            exit(0);
        }

        $this->_query( 'DROP PROCEDURE `REMOVE_NODE_MP`');
        $this->_query( '--
-- ПРОЦЕДУРА УДАЛЕНИЯ ВЕТКИ В ДЕРЕВЕ ТИПА MATERIALIZED PATH
--
-- @param node_id - идентификатор узла, который следует удалить
--
CREATE PROCEDURE REMOVE_NODE_MP( node_id BIGINT)
BEGIN
  DECLARE n_pos, ch_id, p_id, ch_pos BIGINT;
  DECLARE n_path, ch_path, p_path VARCHAR(100);
  DECLARE done, p_cnt, p_rows INT DEFAULT 0;

  -- Создаем курсор, в который вибираем все ветки справа
  -- от перемещаемой, находящиеся с ней на одном уровне
  DECLARE pChildren CURSOR FOR
    SELECT SQL_CALC_FOUND_ROWS node.id, node.path,
      CAST(SUBSTRING(REVERSE(node.path), 1, LOCATE(\'.\', node.path)-1) AS UNSIGNED) as pos
      FROM mp_tree node, mp_tree parent
      WHERE node.path LIKE CONCAT(parent.path, \'%\') AND node.level = parent.level + 1
      AND CAST(SUBSTRING(REVERSE(node.path), 1, LOCATE(\'.\', node.path)-1) AS UNSIGNED) > n_pos
      AND parent.id = p_id
    ORDER BY pos;

  DECLARE CONTINUE HANDLER FOR NOT FOUND SET done = 1;

  -- Собираем необходимые данные
  SELECT path, CAST(SUBSTRING(REVERSE(path), 1, LOCATE(\'.\', path)-1) AS UNSIGNED) INTO n_path, n_pos FROM mp_tree
  WHERE id = node_id;

  SELECT parent.id, parent.path INTO p_id, p_path FROM mp_tree node, mp_tree parent
  WHERE parent.path = SUBSTRING( node.path, 1, (LENGTH(node.path) - LOCATE(\'.\', REVERSE(node.path))))
  AND node.id = node_id;

  -- Удаляем вветку
  DELETE FROM mp_tree WHERE path LIKE CONCAT( n_path, \'%\');

  -- Исправляем данные о путях в ветке из удалили заданную ветку
  OPEN pChildren;
  SELECT FOUND_ROWS() INTO p_rows;

  WHILE p_cnt < p_rows DO
    FETCH pChildren INTO ch_id, ch_path, ch_pos;
    UPDATE mp_tree SET path = REPLACE_PATH(path, ch_path, CONCAT(p_path, \'.\', ch_pos - 1)) WHERE path LIKE CONCAT( ch_path, \'%\');
    SET p_cnt = p_cnt + 1;
  END WHILE;
  CLOSE pChildren;
END');
        if (mysql_errno()) {
            echo "\n" . CLR_LIGHTRED . 'DATABASE ERROR: ' . mysql_error() . CLR_NONE . "\n";
            exit(0);
        }
        echo CLR_LIGHTGREEN . " Done!\n" . CLR_NONE;
    }

    private function _fixDbTables() {
        echo CLR_WHITE . "Prepearing database... " . CLR_NONE;
        $result = $this->_query( "SHOW TABLES");
        if ($result) {
            while ($table = mysql_fetch_array( $result)) {
                $name = $table[0];
                if (isset( $this->_tables[$name])) {
                    $this->_tables[$name][0] = true;
                }
            }
        }
        foreach ($this->_tables as $name => $data) {
            if ($data[0]) {
                $this->_query( "TRUNCATE TABLE `$name`");
            } else {
                $this->_query( $data[1]);
            }
        }
        echo CLR_LIGHTGREEN . "Done!\n" . CLR_NONE;
    }

    private function _saveAL() {
        echo CLR_WHITE . "Building SQL...\n" . CLR_NONE;
        $start = microtime( true);
        $tree = &$this->_tree;
        $sql = "INSERT INTO al_tree VALUES\n";

        for ($i = 0, $s = sizeof( $tree); $i < $s; $i++) {
            $parent = $i == 0 ? 'NULL' : $tree[$i][1] + 1;
            $name = $tree[$i][0];
            $id = $i + 1;
            $sql .= "($id, $parent, '$name')";
            $sql .= ($i < $s - 1) ? ',' : ';';
            $sql .= "\n";
        }

        $time = number_format( microtime( true) - $start, 2, '.', ' ');
        echo CLR_LIGHTGREEN . "SQL for AL-Tree built in $time seconds.\n" . CLR_NONE;
        echo CLR_WHITE . "Executing SQL...\n" . CLR_NONE;
        $start = microtime( true);
        $this->_query( $sql);
        if (mysql_errno()) {
            echo CLR_LIGHTRED . 'DATABASE ERROR: ' . mysql_error() . "\n" . CLR_NONE;
        } else {
            $time = number_format( microtime( true) - $start, 2, '.', ' ');
            echo CLR_LIGHTGREEN . "SQL executed in $time seconds!\n" . CLR_NONE;
        }
    }

    private function _buildAlTree() {
        $tree = array();
        echo CLR_WHITE . "Building random AL-Tree in memory...\n" . CLR_NONE;
        $start = microtime( true);
        $this->_treeize();
        $time = number_format( microtime( true) - $start, 2, '.', ' ');
        $size = number_format( sizeof( $this->_tree), 0, '', "'");
        echo CLR_LIGHTGREEN . "AL-Tree of $size nodes built in $time seconds!\n" . CLR_NONE;
        $this->_saveAL();
        $this->_tree = null;
        unset( $this->_tree);
    }

    private function _treeize() {
        $this->_tree[0] = array(
            $this->_name(),
            null,
            0
        );
        $maxlvl = 0;
        for ($indx = 0; $indx < TREE_WEIGHT; $indx++) {
            $parent = mt_rand( 0, $indx - 1);
            if ($this->_tree[$parent][2] < TREE_HEIGHT) {
                $level = $this->_tree[$parent][2] + 1;
                $this->_tree[$indx] = array(
                    $this->_name(),
                    $parent,
                    $level
                );
                $maxlvl = $level > $maxlvl ? $level : $maxlvl;
            } else {
                $indx--;
            }
        }
        echo "Max level is: $maxlvl\n";
    }

    private function _buildNsTree() {
        echo CLR_WHITE . "Building NS-Tree...\n" . CLR_NONE;
        $start = microtime( true);
        $this->_tree = array();
        $root = mysql_fetch_assoc( $this->_query( 'SELECT * FROM al_tree WHERE parent_id IS NULL'));
        $this->_rkey = 0;
        $this->_indx = -1;
        $this->_rwalkALNS( $root);
        $time = number_format( microtime( true) - $start, 2, '.', ' ');
        $size = number_format( sizeof( $this->_tree), 0, '', "'");
        echo CLR_LIGHTGREEN . "NS-Tree of $size nodes built in $time seconds!\n" . CLR_NONE;
        $this->_saveNS();
        $this->_tree = null;
        unset( $this->_tree);
        unset( $this->_indx);
        unset( $this->_rkey);
    }

    private function _saveNS() {
        echo CLR_WHITE . "Building SQL for NS-Tree...\n" . CLR_NONE;
        $start = microtime( true);
        $sql = "INSERT INTO ns_tree VALUES\n";
        for ($i = 0, $s = sizeof( $this->_tree); $i < $s; $i++) {
            $lft   = (int)$this->_tree[$i]['lft'];
            $rgt   = (int)$this->_tree[$i]['rgt'];
            $id    = (int)$this->_tree[$i]['id'];
            $name  = mysql_escape_string( $this->_tree[$i]['name']);
            $level = (int)$this->_tree[$i]['level'];
            $sql .= "($id, '$name', $lft, $rgt, $level)";
            $sql .= $i == $s - 1 ? ";\n" : ",\n";
        }
        $time = number_format( microtime( true) - $start, 2, '.', ' ');
        echo CLR_LIGHTGREEN . "SQL built in $time seconds!\n" . CLR_NONE;
        echo CLR_WHITE . "Executing NS-Tree SQL...\n" . CLR_NONE;
        $start = microtime( true);
        $this->_query( $sql);
        if (mysql_errno()) {
            echo CLR_LIGHTRED . 'DATABASE ERROR: ' . mysql_error() . "\n" . CLR_NONE;
        } else {
            $time = number_format( microtime( true) - $start, 2, '.', ' ');
            echo CLR_LIGHTGREEN . "SQL executed in $time seconds!\n" . CLR_NONE;
        }
    }

    private function _rwalkALNS( $node, $level = 0) {
        $indx = ++$this->_indx;
        $this->_tree[$indx] = array(
            'id'        => $node['id'],
            'name'      => $node['name'],
            'level'     => $level,
            'lft'       => ++$this->_rkey
        );
        $level++;
        $children = $this->_query( 'SELECT * FROM al_tree WHERE parent_id = ' . (int)$node['id']);
        while ($child = mysql_fetch_assoc( $children)) {
            $this->_rwalkALNS( $child, $level);
        }
        $this->_tree[$indx]['rgt'] = ++$this->_rkey;
    }


    private function _buildMpTree() {
        echo CLR_WHITE . "Building MP-Tree...\n" . CLR_NONE;
        $start = microtime( true);
        $this->_tree = array();
        $this->_indx = -1;
        $root = mysql_fetch_assoc( $this->_query( 'SELECT * FROM al_tree WHERE parent_id IS NULL'));
        $this->_rwalkALMP( $root);
        $time = number_format( microtime( true) - $start, 2, '.', ' ');
        $size = number_format( sizeof( $this->_tree), 0, '', "'");
        echo CLR_LIGHTGREEN . "MP-Tree of $size nodes built in $time seconds!\n" . CLR_NONE;
        $this->_saveMP();
        $this->_tree = null;
        unset( $this->_tree);
        unset( $this->_indx);
    }

    private function _rwalkALMP( $node, $path = '1', $level = 0) {
        $this->_tree[++$this->_indx] = array(
            'id'        => $node['id'],
            'name'      => $node['name'],
            'path'      => $path,
            'level'     => $level
        );
        $level++;
        $children = $this->_query( 'SELECT * FROM al_tree WHERE parent_id = ' . (int)$node['id']);
        $add_path = 0;
        while ($child = mysql_fetch_assoc( $children)) {
            $child_path = $path . '.' . (++$add_path);
            $this->_rwalkALMP( $child, $child_path, $level);
        }
    }

    private function _saveMP() {
        echo CLR_WHITE . "Building SQL for MP-Tree...\n" . CLR_NONE;
        $start = microtime( true);
        $sql = "INSERT INTO mp_tree VALUES\n";
        for ($i = 0, $s = sizeof( $this->_tree); $i < $s; $i++) {
            $id    = (int)$this->_tree[$i]['id'];
            $name  = mysql_escape_string( $this->_tree[$i]['name']);
            $path  = mysql_escape_string( $this->_tree[$i]['path']);
            $level = (int)$this->_tree[$i]['level'];
            $sql .= "($id, '$name', '$path', $level)";
            $sql .= $i == $s - 1 ? ";\n" : ",\n";
        }
        $time = number_format( microtime( true) - $start, 2, '.', ' ');
        echo CLR_LIGHTGREEN . "SQL built in $time seconds!\n" . CLR_NONE;
        echo CLR_WHITE . "Executing MP-Tree SQL...\n" . CLR_NONE;
        $start = microtime( true);
        $this->_query( $sql);
        if (mysql_errno()) {
            echo CLR_LIGHTRED . 'DATABASE ERROR: ' . mysql_error() . "\n" . CLR_NONE;
        } else {
            $time = number_format( microtime( true) - $start, 2, '.', ' ');
            echo CLR_LIGHTGREEN . "SQL executed in $time seconds!\n" . CLR_NONE;
        }
    }

}
