<?php
require_once 'Interface.php';
require_once 'db/Abstract.php';

class TestTask extends Db implements TaskInterface {

    private $_tests = array();
    private $_resfile = '';
    private $_results = null;

    public function __construct() {
        if (!is_dir( './results')) {
            if (!@mkdir( './results')) {
                echo CLR_LIGHTRED . "ERROR: Can not open results directory. Are permissions ok?\n" . CLR_NONE;
            }
        }
        $this->_resfile = './results/' . date( 'Y-m-d-H-i:s') . '.txt';
    }

    public function run() {
        $args = TaskManager::getInstance()->arguments();
        $this->_testModels();
    }

    private function _testModels() {
        echo CLR_WHITE . "SELECT queries testing started...\n" . CLR_NONE;
        echo 'Step 1 ';
        echo ".";
        $this->_testSelectAll();
        echo ". " . CLR_LIGHTGREEN . 'Done!' . CLR_NONE . "\n";
        $cnt = 1;
        for ($i = 2; $i <= TREE_WEIGHT; $i++) {
            echo 'Step ' . ++$cnt . ' ';
            $i = self::_i( $i);
            $this->_testSelectPathById( $i);
            echo ".";
            $this->_testSelectBranchById( $i);
            echo ".";
            $this->_testSelectParentById( $i);
            echo ".";
            $this->_testSelectChildrenById( $i);
            echo ". " . CLR_LIGHTGREEN . 'Done!' . CLR_NONE . "\n";
        }
        echo CLR_WHITE . "SELECT queries testing finished!\n" . CLR_NONE;

        echo CLR_WHITE . "MODIFICATION queries testing started...\n" . CLR_NONE;
        $this->_prepareModDb();
        $this->_testAddNode();
        $this->_testMoveNode();
        $this->_testDeleteNode();
        $this->_freeBackup();
        echo CLR_WHITE . "MODIFICATION queries testing finished!\n" . CLR_NONE;

        $this->_saveResult();
        echo CLR_YELLOW . "Testing finished successfully!\n" . CLR_NONE;
    }

    private static function _i( $i) {
        if     ($i > 100000) $i += 100000;
        elseif ($i > 10000)  $i += 10000;
        elseif ($i > 1000)   $i += 1000;
        elseif ($i > 100)    $i += 100;
        elseif ($i > 10)     $i += 10;
        if ($i > TREE_WEIGHT) $i = TREE_WEIGHT;
        return $i;
    }

    private function _errmsg() {
        if (mysql_errno()) {
           $msg = mysql_error();
           echo "DATABASE ERROR: $msg\n";
           return true;
        }
        return false;
    }

    protected function _query( $sql) {
        $res = parent::_query( $sql);
        $this->_errmsg();
        return $res;
    }

    private function _testSelectAll() {
        // ===== AL:
        $start = microtime(true);
        $sql = 'SELECT SQL_NO_CACHE * FROM al_tree';
        $this->_query( $sql);
        $time = microtime(true) - $start;
        $this->_results['AL']['SELECT']['ALL'] = $time;

        // ===== NS:
        $start = microtime(true);
        $sql = 'SELECT SQL_NO_CACHE * FROM ns_tree';
        $this->_query( $sql);
        $time = microtime(true) - $start;
        $this->_results['NS']['SELECT']['ALL'] = $time;

        // ===== MP:
        $start = microtime(true);
        $sql = 'SELECT SQL_NO_CACHE * FROM mp_tree';
        $this->_query( $sql);
        $time = microtime(true) - $start;
        $this->_results['MP']['SELECT']['ALL'] = $time;
    }

    private function _testSelectPathById( $id) {
        // ===== AL:
        $fields = '';
        $join = '';
        for ($i = 1; $i <= TREE_HEIGHT; $i++) {
            $fields .= "t$i.id AS id1, t$i.name AS name1, t$i.parent_id as parent_id$i";
            $fields .= $i == TREE_HEIGHT ? "\n" : ",\n";
            if ($i > 1) {
                $join .= "LEFT JOIN al_tree AS t$i ON t$i.parent_id = t" . ($i - 1) . ".id\n";
            }
        }
        $sql = "SELECT SQL_NO_CACHE $fields FROM al_tree AS t1 $join WHERE t1.parent_id IS NULL AND t" . ($i - 1) . ".id = $id";
        $start = microtime(true);
        $this->_query( $sql);
        $time = microtime(true) - $start;
        $this->_results['AL']['SELECT']['PATH'][$id] = $time;

        // ===== NS:
        $sql = "SELECT SQL_NO_CACHE parent.*
        FROM ns_tree node, ns_tree parent
        WHERE node.lft BETWEEN parent.lft AND parent.rgt
        AND node.id = $id";
        $start = microtime(true);
        $this->_query( $sql);
        $time = microtime(true) - $start;
        $this->_results['NS']['SELECT']['PATH'][$id] = $time;

        // ===== MP
        $sql = "SELECT SQL_NO_CACHE parent.*
        FROM mp_tree parent, mp_tree node
        WHERE node.path LIKE CONCAT(parent.path, '%')
        AND node.id = $id";
        $start = microtime(true);
        $this->_query( $sql);
        $time = microtime(true) - $start;
        $this->_results['MP']['SELECT']['PATH'][$id] = $time;
    }

    private function _selectBranchRAl( $id, $bid) {
        if (!isset( $this->_results['AL']['SELECT']['BRANCH'])) {
            $this->_results['AL']['SELECT']['BRANCH'][$bid] = 0.0;
        }
        $sql = "SELECT SQL_NO_CACHE * FROM al_tree WHERE parent_id = $id";
        $start = microtime(true);
        $children = $this->_query( $sql);
        $time = microtime(true) - $start;
        $this->_results['AL']['SELECT']['BRANCH'][$bid] += $time;
        if ($children) {
            while ($child = mysql_fetch_assoc( $children)) {
                $this->_selectBranchRAl( $child['id'], $bid);
            }
        }
    }

    private function _testSelectBranchById( $id) {
        // ===== AL
        $this->_selectBranchRAl( $id, $id);

        // ===== NS
        $sql = "SELECT SQL_NO_CACHE node.*
        FROM ns_tree node, ns_tree parent
        WHERE node.lft >= parent.lft AND node.rgt <= parent.rgt
        AND parent.id = $id";
        $start = microtime(true);
        $this->_query( $sql);
        $time = microtime(true) - $start;
        $this->_results['NS']['SELECT']['BRANCH'][$id] = $time;

        // ===== MP
        $sql = "SELECT SQL_NO_CACHE node.*
        FROM mp_tree node, mp_tree parent
        WHERE node.path LIKE CONCAT(parent.path, '%')
        AND parent.id = $id";
        $start = microtime(true);
        $this->_query( $sql);
        $time = microtime(true) - $start;
        $this->_results['MP']['SELECT']['BRANCH'][$id] = $time;
    }

    private function _testSelectParentById( $id) {
        // ===== AL
        $sql = "SELECT SQL_NO_CACHE parent.*
        FROM al_tree node, al_tree parent
        WHERE node.parent_id = parent.id
        AND node.id = $id";
        $start = microtime(true);
        $this->_query( $sql);
        $time = microtime(true) - $start;
        $this->_results['AL']['SELECT']['PARENT'][$id] = $time;

        // ===== NS
        $sql = "SELECT SQL_NO_CACHE parent.*
        FROM ns_tree node, ns_tree parent
        WHERE node.lft BETWEEN parent.lft AND parent.rgt
        AND parent.level = node.level - 1 AND node.id = $id";
        $start = microtime(true);
        $this->_query( $sql);
        $time = microtime(true) - $start;
        $this->_results['NS']['SELECT']['PARENT'][$id] = $time;

        // ===== MP
        $sql = "SELECT SQL_NO_CACHE parent.*
        FROM mp_tree node, mp_tree parent
        WHERE parent.path = SUBSTRING( node.path, 1, (LENGTH(node.path) - LOCATE('.', REVERSE(node.path))))
        AND node.id = $id";
        $start = microtime(true);
        $this->_query( $sql);
        $time = microtime(true) - $start;
        $this->_results['MP']['SELECT']['PARENT'][$id] = $time;
    }

    private function _testSelectChildrenById( $id) {
        // ===== AL
        $sql = "SELECT SQL_NO_CACHE * FROM al_tree WHERE parent_id = $id";
        $start = microtime(true);
        $this->_query( $sql);
        $time = microtime(true) - $start;
        $this->_results['AL']['SELECT']['CHILDREN'][$id] = $time;

        // ===== NS
        $sql = "SELECT SQL_NO_CACHE node.*
        FROM ns_tree node, ns_tree parent
        WHERE node.lft BETWEEN parent.lft AND parent.rgt
        AND node.level = parent.level + 1 and parent.id = $id";
        $start = microtime(true);
        $this->_query( $sql);
        $time = microtime(true) - $start;
        $this->_results['NS']['SELECT']['CHILDREN'][$id] = $time;

        // ===== MP
        $sql = "SELECT SQL_NO_CACHE node.*
        FROM mp_tree node, mp_tree parent
        WHERE node.path LIKE CONCAT(parent.path, '%')
        AND node.level = parent.level + 1
        AND parent.id = $id";
        $start = microtime(true);
        $this->_query( $sql);
        $time = microtime(true) - $start;
        $this->_results['MP']['SELECT']['CHILDREN'][$id] = $time;
    }

    private function _prepareModDb() {
        echo CLR_WHITE . "Backing up data before make changes";
        $this->_query( 'DROP TEMPORARY TABLE IF EXISTS mod_al_tree');
        $this->_query( 'CREATE TEMPORARY TABLE mod_al_tree LIKE al_tree');
        $this->_query( 'INSERT INTO mod_al_tree SELECT * FROM al_tree');
        echo ".";
        $this->_query( 'DROP TEMPORARY TABLE IF EXISTS mod_ns_tree');
        $this->_query( 'CREATE TEMPORARY TABLE mod_ns_tree LIKE ns_tree');
        $this->_query( 'INSERT INTO mod_ns_tree SELECT * FROM ns_tree');
        echo ".";
        $this->_query( 'DROP TEMPORARY TABLE IF EXISTS mod_mp_tree');
        $this->_query( 'CREATE TEMPORARY TABLE mod_mp_tree LIKE mp_tree');
        $this->_query( 'INSERT INTO mod_mp_tree SELECT * FROM mp_tree');
        echo ". " . CLR_NONE;
        echo CLR_LIGHTGREEN . "Done!\n" . CLR_NONE;
    }

    private function _testAddNode() {
        echo CLR_WHITE . "INSERT new nodes testing started...\n". CLR_NONE;
        $parent_id = 2; // Choosed as one of the most bigger branches
        $name = $this->_name();

        echo "AL-Tree.";
        $sql = "INSERT INTO al_tree VALUES (0, $parent_id, '$name')";
        $start = microtime(true);
        $this->_query( $sql);
        $time = microtime(true) - $start;
        $this->_results['AL']['INSERT'] = $time;
        echo "." . CLR_LIGHTGREEN . "Done!\n" . CLR_NONE;

        echo "NS-Tree.";
        $start = microtime(true);
        $this->_query( "BEGIN");
        $this->_query( "SELECT @treeRight := rgt FROM ns_tree WHERE id = $parent_id");
        $this->_query( "SELECT @nodeLevel := parent.level + 1 FROM ns_tree node, ns_tree parent
        WHERE node.lft BETWEEN parent.lft AND parent.rgt AND parent.level = node.level - 1 AND node.id = $parent_id");
        $this->_query( "UPDATE ns_tree SET rgt = rgt + 2 WHERE rgt > @treeRight");
        $this->_query( "UPDATE ns_tree SET lft = lft + 2 WHERE lft > @treeRight");
        $this->_query( "INSERT INTO ns_tree VALUES(0, '$name', @treeRight + 1, @treeRight + 2, @nodeLevel)");
        $this->_query( "COMMIT");
        $time = microtime(true) - $start;
        $this->_results['NS']['INSERT'] = $time;
        echo "." . CLR_LIGHTGREEN . "Done!\n" . CLR_NONE;

        echo "MP-Tree.";
        $start = microtime(true);
        $this->_query( "BEGIN");
        $this->_query( "SELECT @childPos := count(node.id) + 1
        FROM mp_tree node, mp_tree parent
        WHERE node.path LIKE CONCAT(parent.path, '%') AND node.level = parent.level + 1 AND parent.id = $parent_id");
        $this->_query( "SELECT @nodePath := CONCAT(parent.path, '.', @childPos), @nodeLevel := parent.level + 1
        FROM mp_tree node, mp_tree parent
        WHERE parent.path = SUBSTRING( node.path, 1, (LENGTH(node.path) - LOCATE('.', REVERSE(node.path)))) AND node.id = $parent_id");
        $this->_query( "INSERT INTO mp_tree VALUES(0, '$name', @nodePath, @nodeLevel)");
        $this->_query( "COMMIT");
        $time = microtime(true) - $start;
        $this->_results['MP']['INSERT'] = $time;
        echo "." . CLR_LIGHTGREEN . "Done!\n" . CLR_NONE;

        $this->_rollbackChanges();
        echo CLR_LIGHTGREEN . "INSERT new nodes testing finished!\n" . CLR_NONE;
    }

    private function _testMoveNode() {
        echo CLR_WHITE . "UPDATE (move) nodes testing started...\n". CLR_NONE;

        $id = 2;
        $p = mysql_fetch_assoc( $this->_query( "SELECT id FROM ns_tree WHERE id NOT IN (SELECT node.id FROM ns_tree node, ns_tree parent WHERE node.lft >= parent.lft AND node.rgt <= parent.rgt AND parent.id = 2) AND id > 1 LIMIT 0,1"));
        $pid = $p['id'];

        echo "AL-Tree.";
        $sql = "UPDATE al_tree SET parent_id = $pid WHERE id = $id;";
        $start = microtime(true);
        $this->_query( $sql);
        $time = microtime(true) - $start;
        $this->_results['AL']['UPDATE'] = $time;
        echo "." . CLR_LIGHTGREEN . "Done!\n" . CLR_NONE;

        echo "NS-Tree.";
        $sql = "CALL MOVE_NODE_NS( $id, $pid)";
        $start = microtime(true);
        $this->_query( $sql);
        $time = microtime(true) - $start;
        $this->_results['NS']['UPDATE'] = $time;
        echo "." . CLR_LIGHTGREEN . "Done!\n" . CLR_NONE;

        echo "MP-Tree.";
        $sql = "CALL MOVE_NODE_MP( $id, $pid)";
        $start = microtime(true);
        $this->_query( $sql);
        $time = microtime(true) - $start;
        $this->_results['MP']['UPDATE'] = $time;
        echo "." . CLR_LIGHTGREEN . "Done!\n" . CLR_NONE;

        $this->_rollbackChanges();
        echo CLR_LIGHTGREEN . "UPDATE (move) nodes testing finished!\n" . CLR_NONE;
    }

    private function _testDeleteNode() {
        echo CLR_WHITE . "DELETE nodes testing started...\n". CLR_NONE;
        $data = mysql_fetch_assoc( $this->_query( "SELECT MAX(level) as maxlvl, ROUND(MAX(level)/2) as limstart FROM ns_tree"));
        $maxlvl = $data['maxlvl'];
        $limstart = $data['limstart'];
        $id = $this->_query( "SELECT parent.id as id
        FROM ns_tree node, ns_tree parent
        WHERE node.lft BETWEEN parent.lft AND parent.rgt
        AND node.id = (SELECT id FROM ns_tree WHERE level = $maxlvl LIMIT 0,1)
        ORDER BY id DESC LIMIT $limstart,1");
        $id = (int)$id['id'];

        echo "AL-Tree.";
        $sql = "DELETE FROM al_tree WHERE id = $id";
        $start = microtime(true);
        $this->_query( $sql);
        $time = microtime(true) - $start;
        $this->_results['AL']['DELETE'] = $time;
        echo "." . CLR_LIGHTGREEN . "Done!\n" . CLR_NONE;

        echo "NS-Tree.";
        $start = microtime(true);
        $this->_query( "BEGIN");
        $this->_query( "SELECT @treeLeft := lft, @treeRight := rgt, @treeWidth := rgt - lft + 1 FROM ns_tree WHERE id = $id");
        $this->_query( "DELETE FROM ns_tree WHERE lft BETWEEN @treeLeft AND @treeRight");
        $this->_query( "UPDATE ns_tree SET rgt = rgt - @treeWidth WHERE rgt > @treeRight");
        $this->_query( "UPDATE ns_tree SET lft = lft - @treeWidth WHERE lft > @treeRight");
        $this->_query( "COMMIT");
        $time = microtime(true) - $start;
        $this->_results['NS']['DELETE'] = $time;
        echo "." . CLR_LIGHTGREEN . "Done!\n" . CLR_NONE;

        echo "MP-Tree.";
        $start = microtime(true);
        $this->_query( "CALL REMOVE_NODE_MP($id)");
        $time = microtime(true) - $start;
        $this->_results['MP']['DELETE'] = $time;
        echo "." . CLR_LIGHTGREEN . "Done!\n" . CLR_NONE;

        $this->_rollbackChanges();
        echo CLR_LIGHTGREEN . "DELETE nodes testing finished!\n" . CLR_NONE;
    }

    private function _rollbackChanges() {
        echo CLR_WHITE . "Rolling back changes";
        $this->_query( 'TRUNCATE TABLE al_tree');
        $this->_query( 'INSERT INTO al_tree SELECT * FROM mod_al_tree');
        echo ".";
        $this->_query( 'TRUNCATE TABLE ns_tree');
        $this->_query( 'INSERT INTO ns_tree SELECT * FROM mod_ns_tree');
        echo ".";
        $this->_query( 'TRUNCATE TABLE mp_tree');
        $this->_query( 'INSERT INTO mp_tree SELECT * FROM mod_mp_tree');
        echo ". " . CLR_NONE;
        echo CLR_LIGHTGREEN . "Done!\n" . CLR_NONE;
    }

    private function _freeBackup() {
        $this->_query( 'DROP TEMPORARY TABLE mod_al_tree');
        $this->_query( 'DROP TEMPORARY TABLE mod_ns_tree');
        $this->_query( 'DROP TEMPORARY TABLE mod_mp_tree');
    }

    private function _saveResult() {
        echo CLR_YELLOW . "Saving results... " . CLR_NONE;
        $dir = './results/' . date( 'Y-m-d') . '/' . date( 'H-i-s');
        if (!$this->_mkpath( $dir)) {
            echo CLR_LIGHTRED . "ERROR: cannot open $dir directory. Are permissions OK?\n" . CLR_NONE;
        }
        $weight = mysql_fetch_assoc( $this->_query( "SELECT COUNT(*) as w FROM ns_tree"));
        $weight = $weight['w'];
        $height = mysql_fetch_assoc( $this->_query( "SELECT MAX(level) + 1 as h FROM ns_tree"));
        $height = $height['h'];
        file_put_contents( "$dir/README.TXT", "Количество элементов в дереве: $weight\nВысота дерева: $height\n");
        foreach ($this->_results as $type => $data) {
            file_put_contents( "$dir/$type.csv", $this->_getCsvData( $data));
        }
        echo CLR_LIGHTGREEN . "Done!\n" . CLR_NONE;
    }

    private function _getCsvData( $data) {
        $title = 'Operation';
        $body = '';
        $set = false;
        $cnt = sizeof($data['SELECT']['PATH']);
        $body .= 'SELECT ALL TREE;' . $data['SELECT']['ALL'] . str_repeat( ';', $cnt-1) . "\n";
        foreach ($data['SELECT'] as $op => $results) {
            if ($op != 'ALL') {
                $i = 0;
                $body .= 'SELECT ' . $op;
                foreach ($results as $result) {
                    if (!$set) $title .= ';' . ++$i;
                    $body .= ';' . $result;
                }
                $body .= "\n";
                $set = true;
            }
        }
        $body .= 'INSERT (ADD NODE);' . $data['INSERT'] . str_repeat( ';', $cnt-1) . "\n";
        $body .= 'UPDATE (MOVE NODE);' . $data['UPDATE'] . str_repeat( ';', $cnt-1) . "\n";
        $body .= 'DELETE (REMOVE NODE);' . $data['DELETE'] . str_repeat( ';', $cnt-1) . "\n";
        return $title . "\n" . $body;
    }

    private function _mkpath( $path) {
        if(@mkdir( $path) or file_exists( $path)) return true;
        return ($this->_mkpath( dirname( $path)) and mkdir( $path));
    }
}
